/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * datafile.c: Manages loading and saving files
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#include "common.h"
#include "datafile.h"

#include "zlib/zlib.h"

uint8 readu8(buffer * buf) {
    if (buf->pos + 1 > buf->len)
	return 0;
    return buf->data[buf->pos++];
}

uint16 readu16(buffer * buf) {
    uint16 ret;
    if (buf->pos + 2 > buf->len)
	return 0;
    ret = buf->data[buf->pos++] << 8;
    ret |= buf->data[buf->pos++];
    return ret;
}

uint32 readu32(buffer * buf) {
    uint32 ret;
    if (buf->pos + 4 > buf->len)
	return 0;
    ret = buf->data[buf->pos++] << 24;
    ret |= buf->data[buf->pos++] << 16;
    ret |= buf->data[buf->pos++] << 8;
    ret |= buf->data[buf->pos++];
    return ret;
}

uint16 readu16le(buffer * buf) {
    uint16 ret;
    if (buf->pos + 2 > buf->len)
	return 0;
    ret = buf->data[buf->pos++];
    ret |= buf->data[buf->pos++] << 8;
    return ret;
}

uint32 readu32le(buffer * buf) {
    uint32 ret;
    if (buf->pos + 4 > buf->len)
	return 0;
    ret = buf->data[buf->pos++];
    ret |= buf->data[buf->pos++] << 8;
    ret |= buf->data[buf->pos++] << 16;
    ret |= buf->data[buf->pos++] << 24;
    return ret;
}

void readstr(buffer * buf, char* ret, int bl) {
    uint16 l = readu16(buf);
    uint16 readl = l;
    uint16 copyl = l;
    bl--;
    if (copyl > bl) copyl = bl;
    if (readl + buf->pos > buf->len) readl = buf->len - buf->pos;
    if (copyl > readl) copyl = readl;

    if (copyl)
	RB memcpy(ret,&buf->data[buf->pos],copyl);
    ret[copyl] = 0;
    buf->pos += readl;
}
int readstr_alloc(buffer * buf, char** ret) {
    uint16 l = readu16(buf);
    uint16 readl = l;
    if (readl + buf->pos > buf->len) readl = buf->len - buf->pos;

    *ret = (char*)malloc(readl+1);

    if (readl)
	RB memcpy(*ret,&buf->data[buf->pos],readl);
    (*ret)[readl] = 0;
    buf->pos += readl;
    return readl;
}

void writeu8(buffer* buf, uint8 v) {
    expand(buf->data, buf->cap, buf->pos + 1);
    buf->data[buf->pos++] = v;
    if (buf->len < buf->pos) buf->len = buf->pos;
}
void writeu16(buffer* buf, uint16 v) {
    expand(buf->data, buf->cap, buf->pos + 2);
    buf->data[buf->pos++] = v>>8;
    buf->data[buf->pos++] = v;
    if (buf->len < buf->pos) buf->len = buf->pos;
}
void writeu32(buffer* buf, uint32 v) {
    expand(buf->data, buf->cap, buf->pos + 4);
    buf->data[buf->pos++] = v>>24;
    buf->data[buf->pos++] = v>>16;
    buf->data[buf->pos++] = v>>8;
    buf->data[buf->pos++] = v;
    if (buf->len < buf->pos) buf->len = buf->pos;
}

void writestr(buffer* buf, const char* str) {
    writebuf(buf, (uint8*)str, RB strlen(str));
}
void writedata(buffer* buf, uint8* data, int l) {
    expand(buf->data, buf->cap, buf->pos + l);
    RB memcpy(&buf->data[buf->pos], data, l);
    buf->pos += l;
    if (buf->len < buf->pos) buf->len = buf->pos;
}
void writebuf(buffer* buf, uint8* data, int l) {
    writeu16(buf, l);
    writedata(buf, data, l);
}

void serial_u8(uint8* p, buffer* buf, int phase) {
    if (phase == PHASE_WRITE) {
        writeu8(buf, *p);
    } else if (phase == PHASE_READ) {
        *p = readu8(buf);
    }
}
void serial_u16(uint16* p, buffer* buf, int phase) {
    if (phase == PHASE_WRITE) {
        writeu16(buf, *p);
    } else if (phase == PHASE_READ) {
        *p = readu16(buf);
    }
}
void serial_u32(uint32* p, buffer* buf, int phase) {
    if (phase == PHASE_WRITE) {
        writeu32(buf, *p);
    } else if (phase == PHASE_READ) {
        *p = readu32(buf);
    }
}
void serial_str(char** s, buffer* buf, int phase) {
    if (phase == PHASE_WRITE) {
        writestr(buf, *s);
    } else if (phase == PHASE_READ) {
        readstr_alloc(buf, s);
    }
}

void serial_buffer(buffer* b2, buffer* buf, int phase) {
    if (phase == PHASE_WRITE) {
        writebuf(buf, b2->data, b2->len);
    } else if (phase == PHASE_READ) {
        b2->len = readstr_alloc(buf, (char**)&b2->data);
        b2->pos = 0;
        b2->cap = b2->len + 1;
    }
}


#ifdef SDLTEST
size_t get_file_size(int fd) {
    off_t cpos = lseek(fd,0,SEEK_CUR);
    off_t size = lseek(fd,0,SEEK_END);
    lseek(fd,cpos,SEEK_SET);
    return size;
}
#else
#define get_file_size(fd) RB filesize(fd)
#endif
    


int loadfile(const char* fname, buffer* buf) {
    int fd;

    fd = RB open(fname, O_RDONLY);
    if (fd < 0) {
	return -1;
    }

    buf->pos = 0;
    buf->len = get_file_size(fd);
    buf->data = malloc(buf->len);

    if (!buf->data) {
	return -2;
    }
    RB read(fd,buf->data,buf->len);
    RB close(fd);

    return 0;
}

int writefile(const char* fname, buffer* buf) {
    int fd;

#ifdef SDLTEST
    fd = RB open(fname, O_WRONLY | O_CREAT | O_TRUNC, 0644);
#else
    fd = RB open(fname, O_WRONLY | O_CREAT | O_TRUNC);
#endif
    if (fd < 0) {
	return -1;
    }

    RB write(fd, buf->data, buf->len);
    RB close(fd);

    return 0;
}

int loadgzfile(const char* fname, buffer* buf) {
    gzFile f;

    f = gzopen(fname, "r");
    if (!f) {
	return -1;
    }

    buf->len = 24576;
    buf->data = malloc(buf->len);
    if (!buf->data) {
	gzclose(f);
	return -2;
    }

    buf->pos = 0;
    while (1) {
	int len = gzread(f,&buf->data[buf->pos],
			 buf->len - buf->pos);
	buf->pos += len;
	if (len < 0) {
	    gzclose(f);
	    free(buf->data);
	    return -1;
	}
	if (len == 0) break;
	buf->len *= 3;
	buf->len >>= 1;
	buf->data = realloc(buf->data, buf->len);
	if (!buf->data) {
	    gzclose(f);
	    return -2;
	}

    }
    buf->len = buf->pos;
    buf->data = realloc(buf->data, buf->len);
    buf->pos = 0;

    gzclose(f);
    return 0;
}
