/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * imath.h: Various math functions
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#ifndef __IMATH_H_
#define __IMATH_H_

#include "common.h"

#ifdef CPU_COLDFIRE

/*
 * Use the EMAC for faster multiplies.
 */


/* Calculate (x1 * x2 + y1 * y2) >> 16 */

static inline sint32 dotprod(sint32 x1, sint32 y1, sint32 x2, sint32 y2) {
    uint32 lo, hi;			
    asm (" | dotprod \n\t" 
	 "move.l #0, %%macsr\n\t"
	  "mac.l %[x1], %[x2], %%acc0 \n\t"
	  "mac.l %[y1], %[y2], %%acc0 \n\t"
	  "move.l %%accext01, %[hi]    \n\t"
	  "movclr.l %%acc0, %[lo]    \n\t"
	  : [lo] "=r" (lo), [hi] "=r" (hi)
	  : [x1] "r" (x1), [y1] "r" (y1),
	    [x2] "r" (x2), [y2] "r" (y2));

    return (sint32)(hi<<16 | lo>>16);
}

/* Calculate (a * b) >> 16 */
static inline sint32 fixedmul(sint32 a, sint32 b) {
    uint32 lo=b, hi;

    asm ( " | fixedmul \n\t" 
	  "move.l #0x20, %%macsr      \n\t"
          "mac.l %[a], %[lo], %%acc0 \n\t"
	  "movclr.l %%acc0, %[hi]     \n\t"
	  "muls.l  %[a], %[lo]       \n\t"
	  : [hi] "=d" (hi), [lo] "+d" (lo),
	    [a] "+d" (a) );

    return (hi<<15 | lo>>16);

}

/* Calculate (a * b) >> 31 */
static inline sint32 fracmul(sint32 a, sint32 b) {
    uint32 hi;

    asm ( " | fracmul \n\t" 
	  "move.l #0x20, %%macsr      \n\t"
          "mac.l %[a], %[lo], %%acc0 \n\t"
	  "movclr.l %%acc0, %[hi]     \n\t"
	  : [hi] "=d" (hi), [lo] "+d" (b),
	    [a] "+d" (a) );

    return hi;

}

/* Calculate (a * b) 32x32 -> 64*/
static inline sint64 mul64 (sint32 a, sint32 b) {
    union {
	struct {
	    uint32 hi, lo;
	};
	sint64 ret;
    } r;
    r.lo = b;

    asm (" | mul64 \n\t" 
	 "move.l #0x20, %%macsr      \n\t"
          "mac.l %[a], %[lo], %%acc0 \n\t"
	  "movclr.l %%acc0, %[hi]     \n\t"
	  "asr.l  #1,  %[hi]       \n\t"
	  "muls.l  %[a], %[lo]       \n\t"
	  : [hi] "=d" (r.hi), [lo] "+d" (r.lo),
	    [a] "+d" (a) );
    
    return r.ret;
}


/* Most Coldfires don't have a divide instruction, but the MCF5249
   does. Older versions of GCC doesn't know about it.
*/

/* Calculate a / b or saturate if b is 0*/
static inline sint32 div32(sint32 a, sint32 b) {
    asm (" | div32 \n\t" 
         "tst %1\n"
	 "beqs 1f\n"
	 "divs.l %1, %0\n" 
	 "bras 2f\n"
	 "1:\n"
	 "move.l #0x7fffffff, %0 \n" /* saturate */
	 "2:\n"
	 : "+d" (a) : "d" (b));
    return a;
}

#else

/* Generic C versions of the above functions */
#define dotprod(x1,y1,x2,y2) ((((sint64)(x1) * (x2)) + ((sint64)(y1) * (y2)))>>16)
#define fixedmul(a,b) (((sint64)(a) * (b))>>16)
#define mul64(a,b) ((sint64)(a) * (b))
#define div32(a,b) ({ sint32 ___b = (b); ___b == 0 ? 0x7fffffff : ((a) / ___b); })
#define fracmul(a,b) (((sint64)(a) * (b))>>31)

#endif


/* Divide 'a' by a constant using fracmul. This will work if c is not
   constant, but it has absolutely no advantage over div32. */
#define divconst(a,c) fracmul(a,((1u<<31)/((unsigned)c)))

/* Angles go from 0 to 2048 */

extern uint16 sintable[512];

#define icos(t) (isin((t)+512))

static inline int isin(int t) {
    int out, index;
    index = t & 511;
    if (index) {
	if (t & 512) {
	    index = -index;
	    index &= 511;
	}
	out = sintable[index];
    } else {
	out = (t & 512) << 7;
    }
    if (t & 1024)
	out = -out;
    return out;
}

static inline uint32 isqrt(uint32 i) {
    uint32 root, l1, l2;

    root = (i + 1) >> 1;
    l1 = l2 = 0;
    while (root != l1 && root != l2) {
	l2 = l1;
	l1 = root;
	root = (div32(i,root) + root) >> 1;
    }
    return root;
}

static inline uint64 lisqrt(uint64 i) {
    uint64 root, l1, l2;

    root = (i + 1) >> 1;
    l1 = l2 = 0;
    while (root != l1 && root != l2) {
	l2 = l1;
	l1 = root;
	root = (i / root + root) >> 1;
    }
    return root;
}

#endif
