/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * rocken.c: A simple Ken's Labyrinth clone for iRiver H1x0
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/


#include "common.h"
#include "graphics.h"
#include "soundutil.h"
#include "board.h"
#include "actor.h"
#include "player.h"
#include "zlib/zlib.h"
#include "imath.h"
#include "tag.h"
#include "util.h"

PLUGIN_HEADER 

struct plugin_api* rb IBSS_ATTR;
struct plugin_api rb_api IBSS_ATTR;

#define MUS_FADE_SPD 8

/* various constants for menus and sliders*/
#define MENU_TEXT_SIZE 48
#define NUM_SAVE_GAMES 12

#define MENU_TITLE_SPACE_TOP 2
#define MENU_TITLE_SPACE_BOT 1

#define MENU_ITEMS_SPACE_TOP 2
#define MENU_ITEMS_SPACE_BOT 2

#define MENU_ITEMS_SPACE_LEFT 10
#define MENU_ITEMS_SPACE_RIGHT 5

#define MENU_ITEM_PAD_LEFT 2
#define MENU_ITEM_PAD_RIGHT 12
#define MENU_ITEM_PAD_TOP 1
#define MENU_ITEM_PAD_BOT 1

#define MENU_ITEMS_SCROLLBAR_X 3

#define MENU_FONT_TITLE FONT_UI
#define MENU_FONT_ITEMS FONT_SYSFIXED
#define MENU_MIN_WIDTH (LCD_WIDTH/2)

#define SLIDER_WIDTH (LCD_WIDTH*4/5)
#define SLIDER_FONT_TITLE FONT_UI
#define SLIDER_TITLE_SPACE_TOP 2
#define SLIDER_TITLE_SPACE_BOT 1
#define SLIDER_BAR_SPACE_TOP 6
#define SLIDER_BAR_SPACE_BOT 8
#define SLIDER_BAR_SPACE_LEFT 10
#define SLIDER_BAR_SPACE_RIGHT 10

/* menu flags */
#define LEFT_EXITS 1

typedef struct {
    char text[MENU_TEXT_SIZE];
    int res;
} menuitem;

typedef struct {
    char *title;
    int selection;
    int first;
    int nitems;
    int nshown;
    int cap;

    int flags;

    int x, y;
    int w, h;
    int tw, th;
    int iw, ih;
    menuitem* items;
} menu;

typedef struct  {
    int min, max;
    int val;
} slider_data;

typedef enum { A_NONE, A_FWD, A_BACK, A_LEFT, A_RIGHT, 
	       A_SLEFT, A_SRIGHT, A_UP, A_DOWN, A_FIRE
} action;

uint8 keybmp[] = {
    0xf0, 0xfc, 0x7f, 0x1f, 0x1f, 0x3f, 0xfc, 0xf0, 
    0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0x00, 
    0x0f, 0x3f, 0xfd, 0xf0, 0xf0, 0xfc, 0x7f, 0x5f, 
    0x17, 0x03, 0x3f, 0x57, 0x3f, 0x57, 0x3f, 0x55
};

uint8 heartbmp[] = {
    0xf8, 0xfe, 0xfe, 0xf8, 0xf8, 0xfe, 0xfe, 0xf8, 
    0x02, 0x1f, 0x7f, 0xff, 0xff, 0x7f, 0x1f, 0x02
};

int cframe;
static int debug_show_time;
static int boost = 1;

/* cache save game titles so we don't have to spin up the disk to show
   load/save menu
*/
static char save_game_exist[NUM_SAVE_GAMES];
static char save_game_titles[NUM_SAVE_GAMES][32];

OPTION_INT(debug_show_time);

void game_shutdown(void);
void malloc_heapdump(const char*);

int default_button(int button) {
#ifndef SIMULATOR
    switch (button) {
    case (BUTTON_ON|BUTTON_REC):
        malloc_heapdump(PATH "heap.dump");
        RB splash(80, 1, "Heap dumped");
        return 0;
    }
#endif
    return RB default_event_handler_ex(button, (void(*)(void*))game_shutdown, NULL);
}

#define DEBUG_NLINES 8
void debugf(const char *fmt, ...) {
    static char buf[DEBUG_NLINES][64];
    static int cycle = 0;
    int i;

    va_list lst;
    va_start(lst, fmt);
    RB vsnprintf(buf[cycle], 64, fmt, lst);
    va_end(lst);



    RB lcd_remote_setfont(FONT_SYSFIXED);
    RB lcd_remote_clear_display();
    for (i = 0; i < DEBUG_NLINES; i++) {
        cycle++;
        if (cycle >= DEBUG_NLINES) cycle = 0;
        RB lcd_remote_puts(0, i, buf[cycle]);
    }
    RB lcd_remote_update();

    cycle++;
    if (cycle >= DEBUG_NLINES) cycle = 0;

}

/* Create a new menu */
menu* new_menu(char* title, int flags) {
    menu* ret = (menu*) calloc(1, sizeof(menu));

    ret->x = ret->y = -1;
    ret->title = title;
    ret->flags = flags;
    ret->cap = 16;

    ret->items = (menuitem*) malloc(sizeof(menuitem) * ret->cap);

    return ret;
}

void menu_add_item(menu* m, const char* text, int res) {
    expand(m->items, m->cap, m->nitems+1);

    char* txtbuf = m->items[m->nitems].text;

    RB strncpy(txtbuf, text, MENU_TEXT_SIZE);
    txtbuf[MENU_TEXT_SIZE-1] = 0;
    m->items[m->nitems].res = res;
    ++m->nitems;
}

void menu_free(menu* m) {
    free(m->items);
}

void menudraw(menu* m, int active) {
    int bh, i, j, cx, cy;

    if (!m->w) {
        int iw;

        if (!m->nshown) {
            if (m->nitems < 8)
                m->nshown = m->nitems;
            else
                m->nshown = 8;
        }

        RB lcd_setfont(MENU_FONT_TITLE);
        RB lcd_getstringsize(m->title, &m->tw, &m->th);

        RB lcd_setfont(MENU_FONT_ITEMS);

        m->iw = 0;
        for (i = 0; i < m->nitems; i++) {
            RB lcd_getstringsize(m->items[i].text, &iw, &m->ih);
            iw += MENU_ITEM_PAD_LEFT + MENU_ITEM_PAD_RIGHT;
            if (m->iw < iw) m->iw = iw;
        }
        m->ih += MENU_ITEM_PAD_TOP + MENU_ITEM_PAD_BOT;

        m->w = MENU_ITEMS_SPACE_LEFT + m->iw + MENU_ITEMS_SPACE_RIGHT;
        iw = m->tw + 4;
        if (m->w < iw) m->w = iw;
        if (m->w < MENU_MIN_WIDTH) m->w = MENU_MIN_WIDTH;
        m->h = MENU_TITLE_SPACE_TOP + m->th + MENU_TITLE_SPACE_BOT +
            MENU_ITEMS_SPACE_TOP + (m->ih * m->nshown) + MENU_ITEMS_SPACE_BOT;
    }

    if (m->x == -1) 
        m->x = (LCD_WIDTH - m->w) / 2;

    if (m->y == -1) 
        m->y = (LCD_HEIGHT - m->h) / 2;

    bh = MENU_TITLE_SPACE_TOP + m->th + MENU_TITLE_SPACE_BOT;

    RB lcd_set_drawmode(DRMODE_SOLID);
    RB lcd_set_foreground(active ? 2 : 3);
    RB lcd_fillrect(m->x, m->y, m->w, bh);
    RB lcd_set_foreground(0);
    RB lcd_drawrect(m->x, m->y, m->w, bh + 1);


    RB lcd_set_foreground(3);
    RB lcd_fillrect(m->x, m->y + bh, m->w, m->h - bh);
    RB lcd_set_foreground(0);
    RB lcd_drawrect(m->x, m->y + bh, m->w, m->h - bh);

    RB lcd_setfont(MENU_FONT_TITLE);

    RB lcd_set_drawmode(DRMODE_FG);
    RB lcd_set_foreground(active ? 0 : 2);
    RB lcd_putsxy(m->x + (m->w - m->tw) / 2, m->y + MENU_TITLE_SPACE_TOP, m->title);

    if (m->first > m->selection) m->first = m->selection;
    if (m->first < m->selection - (m->nshown-1)) m->first = m->selection - (m->nshown-1);

    cx = m->x + MENU_ITEMS_SPACE_LEFT;
    cy = m->y + bh + MENU_ITEMS_SPACE_TOP;

    RB lcd_set_foreground(0);
    RB scrollbar(m->x + MENU_ITEMS_SCROLLBAR_X, cy, 4, m->ih * m->nshown, 
                  m->nitems, m->first, m->first + m->nshown, VERTICAL);

    RB lcd_set_drawmode(DRMODE_FG);
    RB lcd_setfont(MENU_FONT_ITEMS);
    j = m->first;
    for (i = 0; i < m->nshown && j < m->nitems; i++, j++, cy += m->ih) {
        if (j == m->selection) {
            RB lcd_set_foreground(0);
            RB lcd_fillrect(cx, cy, m->iw, m->ih);
            RB lcd_set_foreground(3);
        } else if (!m->items[j].res) {
            RB lcd_set_foreground(2);
        } else {
            RB lcd_set_foreground(0);
        }
        RB lcd_putsxy(cx + MENU_ITEM_PAD_LEFT, cy + MENU_ITEM_PAD_TOP, m->items[j].text);
    }

    RB lcd_update();

    RB lcd_setfont(FONT_UI);
}

int show_menu(menu* m) {
    int button, j;

    while (1) {
        RB yield();
        menudraw(m, true);
        button = RB button_get(true);

        switch (button & ~BUTTON_REPEAT) {
        case BUTTON_SELECT:
        case BUTTON_RIGHT:
            return m->items[m->selection].res;

        case BUTTON_UP:
            j = m->selection;
            do {
                if (m->selection > 0) {
                    --m->selection;
                } else if (!(button & BUTTON_REPEAT)) {
                    m->selection = m->nitems - 1;
                } else {
                    m->selection = j;
                    break;
                }
            } while (!m->items[m->selection].res);
            break;

        case BUTTON_DOWN:
            j = m->selection;
            do {
                if (m->selection < m->nitems - 1) {
                    ++m->selection;
                } else if (!(button & BUTTON_REPEAT)) {
                    m->selection = 0;
                } else {
                    m->selection = j;
                    break;
                }
            } while (!m->items[m->selection].res);
            break;

        case BUTTON_LEFT:
            if (!(m->flags & LEFT_EXITS))
                break;
        case BUTTON_OFF:
            if (button & BUTTON_REPEAT)
                break;
            return 0;
        case BUTTON_NONE:
            break;
        default:
            if (default_button(button) == SYS_USB_CONNECTED) {
                return MENU_ATTACHED_USB;
            }
            break;
        }

    }
    RB lcd_setfont(FONT_UI);
}

int slider(char* title, void(*func)(void*, slider_data*, int), void* cookie, int savescr) {
    int tw, th;
    int bx, by;
    int bdh, barh;
    int ret = 0;
    int button;
    slider_data dat;
    void* saved = NULL;

    RB lcd_setfont(SLIDER_FONT_TITLE);
    RB lcd_getstringsize(title, &tw, &th);

    bdh = SLIDER_TITLE_SPACE_TOP + th + SLIDER_TITLE_SPACE_BOT;
    barh = SLIDER_BAR_SPACE_TOP + 4 + SLIDER_BAR_SPACE_BOT;

    bx = (LCD_WIDTH - SLIDER_WIDTH) / 2;
    by = (LCD_HEIGHT - (bdh + barh)) / 2;

    if (savescr)
        saved = save_screen();
    while (!ret) {
        func(cookie, &dat, 0);

        RB lcd_set_drawmode(DRMODE_SOLID);
        RB lcd_set_foreground(2);
        RB lcd_fillrect(bx, by, SLIDER_WIDTH, bdh);
        RB lcd_set_foreground(0);
        RB lcd_drawrect(bx, by, SLIDER_WIDTH, bdh + 1);


        RB lcd_set_foreground(3);
        RB lcd_fillrect(bx, by + bdh, SLIDER_WIDTH, barh);
        RB lcd_set_foreground(0);
        RB lcd_drawrect(bx, by + bdh, SLIDER_WIDTH, barh);

        RB lcd_setfont(MENU_FONT_TITLE);

        RB lcd_set_drawmode(DRMODE_FG);
        RB lcd_putsxy((LCD_WIDTH - tw) / 2, by + MENU_TITLE_SPACE_TOP, title);

        RB scrollbar(bx + SLIDER_BAR_SPACE_LEFT, by + bdh + SLIDER_BAR_SPACE_TOP, 
                     SLIDER_WIDTH - (SLIDER_BAR_SPACE_LEFT + SLIDER_BAR_SPACE_RIGHT),
                     4, (dat.max - dat.min), 0, dat.val - dat.min, HORIZONTAL);


        RB lcd_update();


        RB yield();
        button = RB button_get(true);

        switch (button & ~BUTTON_REPEAT) {
        case BUTTON_SELECT:
        case BUTTON_ON:
        case BUTTON_OFF:
            if (!(button & BUTTON_REPEAT)) 
                ret = 1;
            break;
        case BUTTON_LEFT:
        case BUTTON_DOWN:
            func(cookie, &dat, -1);
            break;
        case BUTTON_RIGHT:
        case BUTTON_UP:
            func(cookie, &dat, 1);
            break;

        default:
            if (default_button(button) == SYS_USB_CONNECTED) {
                ret = 3;
            }
            break;
        }
        
    }
    if (saved) {
        restore_screen(saved);
        free(saved);
    }
    return ret;
}

void updatestatus(int time) {
    /* Border */

    RB lcd_set_foreground(0);
    RB lcd_fillrect(0, YDIM, XDIM, LCD_HEIGHT - YDIM);
    RB lcd_set_foreground(2);
    RB lcd_fillrect(1, YDIM + 1, XDIM - 2, LCD_HEIGHT - YDIM - 2);
    RB lcd_set_foreground(3);
    RB lcd_fillrect(2, YDIM + 2, XDIM - 4, LCD_HEIGHT - YDIM - 4);
    RB lcd_set_background(3);

    /* Health */
    if (!(player_get_hurttime(cplayer) & 1)) {

	RB lcd_bitmap(heartbmp, 6, YDIM+4, 8, 8);

	RB lcd_set_foreground(3);
	RB lcd_fillrect(16, YDIM + 3, 24, 10);
	RB lcd_set_foreground(0);
	RB lcd_drawrect(17, YDIM + 4, 22, 8);
	RB lcd_set_foreground(1);
	RB lcd_fillrect(18, YDIM + 5, 2 * player_get_health(cplayer), 6);
    }

    if (HASTAG(cplayer, has_key))
	RB lcd_bitmap(keybmp, 42, YDIM+4, 16, 8);

    RB lcd_setfont(FONT_SYSFIXED);

    if (debug_show_time) {
	char tmp[32];
	int w, h;
	RB snprintf(tmp,32,"%d",time);
	RB lcd_getstringsize(tmp, &w, &h);
	RB lcd_set_foreground(0);
	RB lcd_putsxy(XDIM - 6 - w, YDIM + 4, tmp);
    }

    RB lcd_set_drawmode(DRMODE_SOLID);
    RB lcd_set_foreground(0);

}

void new_game(char* boardname) {
    if (boardname == NULL)
        boardname = PATH "demoboard.rbd";

    object_deref(cplayer);
    cplayer = NULL;
    RB splash(0, 1, "Loading");
    music_change(NULL, 8, 8, false);
    loadboard(boardname);
    
}

void end_game(void) {
    object_deref(cplayer);
    cplayer = NULL;
    loadboard(NULL);
}

void game_shutdown(void) {
    static int shutdown = 0;
    if (!shutdown) {
        shutdown = 1;

        music_change(NULL, MUS_FADE_SPD, 0, false);
        RB sleep(1);

        RB lcd_set_foreground(0);
        RB lcd_set_background(3);
        RB lcd_remote_setfont(FONT_UI);
        RB lcd_setfont(FONT_UI);
        if (cplayer) {
            int tres;
            void* screenbuf = save_screen();
            RB splash(0, 1, "Autosaving...");
            
            tres = save_game(PATH "auto.sav", "Autosave");
            if (tres)
                RB splash(150, 1, "Could not autosave: %d", tres);

            if (screenbuf) {
                restore_screen(screenbuf);
                free(screenbuf);
            }
        }
        
        RB splash(0, 1, "Saving settings...");
        save_settings(PATH "rocken.opt");

        end_game();

#ifndef SIMULATOR
        if (boost)
            RB cpu_boost(0);
#endif

    }

}

int intro_sleep_until(int tks) {
    int ctime, button;
    int endtime = tks;
    while ((ctime = *RB current_tick) < endtime) {
        button = RB button_get_w_tmo(1);
        if (default_button(button) == SYS_USB_CONNECTED) return 3;
        if (!(button & (BUTTON_REL|BUTTON_REPEAT))) {
            switch (button) {
            case BUTTON_ON:
            case BUTTON_OFF:
            case BUTTON_MODE:
            case BUTTON_REC:
            case BUTTON_UP:
            case BUTTON_DOWN:
            case BUTTON_LEFT:
            case BUTTON_RIGHT:
            case BUTTON_SELECT:
                return 1;
            }
        }
        RB yield();
    }
    return 0;
}

TEXGROUP(intro);

void introdraw(int ddist, int ang, int ang2, int tbase, int posz, bool flip, bool cube, bool dbl) {
    int center;
    int s = isin(ang);
    int c = icos(ang);
    int cx, cy;
    /* the middle of the board */
    center = 32<<16 | 32768;

    cx = center - fixedmul(c, 40960+54273+ddist);
    cy = center - fixedmul(s, 40960+54273+ddist);
    
    preparescreen(cx, cy, 0, ang+ang2, 0, 128, 64);
    if (cube) {
        num_sprites = 0;

        addsprite(center-40960, center-32768, 0, intro, 0, DIR_WEST, 0);
        addsprite(center-40960, center+32768, 0, intro, 1, DIR_WEST, 0);
        
        addsprite(center+40960, center+32768, 0, intro, 0, DIR_EAST, 0);
        addsprite(center+40960, center-32768, 0, intro, 1, DIR_EAST, 0);
        
        addsprite(center+32768, center-40960, 0, intro, 0, DIR_NORTH, 0);
        addsprite(center-32768, center-40960, 0, intro, 1, DIR_NORTH, 0);
        
        addsprite(center-32768, center+40960, 0, intro, 0, DIR_SOUTH, 0);
        addsprite(center+32768, center+40960, 0, intro, 1, DIR_SOUTH, 0);

        drawsprites(RB lcd_framebuffer);
    }
    if (tbase != -1) {
        num_sprites = 0;
        if (dbl) {
            addsprite(center-40960, center-32768, posz, intro, tbase+0, DIR_WEST, (flip ? SF_FLIPX : 0));
            addsprite(center-40960, center+32768, posz, intro, tbase+1, DIR_WEST, (flip ? SF_FLIPX : 0));
        } else {
            addsprite(center-40960, center, posz, intro, tbase, DIR_WEST, (flip ? SF_FLIPX : 0));
        }
        drawsprites(RB lcd_framebuffer);
    }
}

int intro(void) {
    int i;
    int res;
    int etime = 0;
    int ang;
    int ang2;
    int ddist = 512;
    int flip = 0;
    int posz = 0;
    int sttime;

    sttime = *RB current_tick;
    
    /* 450*4 = ~17 seconds, the length of the intro music */
    for (i = 0; i < 450; i++) {
        RB lcd_clear_display();

        int ti = i;
        if (ti < 140) {
            if (ti>60) ti = 60;
            ang2 = ((60 - ti) * 220 / 60);
            ang = (ti * 1536 / 60) + 512;
            ang &= 2047;
            flip = ang >= 512 && ang <= 1536;
            introdraw(((60 - ti) * (65536*6)) / 60,
                      -ang, ang2, 4, 0, flip, true, false);
        } else if (ti < 350) {
            ti -= 140;
            if (ti > 30) ti = 30;
            ddist = ti * (ti - 30) * -83;
            ang = ti * 512 / 30;

            introdraw(ddist, -ang, 0, 4, 0, false, true, false);
            introdraw(ddist, -(ang-512), 0, 2, 0, false, false, true);
        } else {
            ti -= 350;
            ang = ti * 2048 / 70;
            ang2 = ti * 256 / 70;
            ddist = ti * (65536*7) / 70;
            ang &= 2047;
            if (ti > 70) ti = 70;
            posz = ti * -24576 / 70;
            flip = ang >= 512 && ang <= 1536;
            introdraw(ddist, -ang, -ang2, -1, 0, false, true, false);
            
            introdraw(0, 0, 0, 2, posz, false, false, true);
        }

        RB lcd_update();
        RB backlight_on();

        etime = sttime + (i*4);
        if ((res = intro_sleep_until(etime))) return res;
        if (!music_is_playing() && i >= 420) return 1;
    }
    return 1;
}

int debugmenu(void) {
    int ret = 0;

    void* screen_save = save_screen();
    menu* m = new_menu("Debug", LEFT_EXITS);

    menu_add_item(m, "Switch CPU frequency", 1);
    menu_add_item(m, "Show time", 2);
    menu_add_item(m, "FSM mode", 3);
    menu_add_item(m, "Heal player", 4);
    menu_add_item(m, "Give key", 5);
    while (!ret) {
        restore_screen(screen_save);
	switch (show_menu(m)) {
	case 1:
#ifndef SIMULATOR
	    boost ^= 1;
	    RB cpu_boost(boost);
	    RB splash(80, 1, "freq: %d", *RB cpu_frequency);
#endif
	    break;
	case 2:
	    debug_show_time ^= 1;
	    if (debug_show_time) {
		RB splash(80, 1, "Time display on");
	    } else {
		RB splash(80, 1, "Time display off");
	    }
	    break;
	case 3: 
	    if (HASTAG(cplayer, is_invulnerable)) {
		RB splash(80, 1, "FSM mode off");
		DELTAG(cplayer, is_invulnerable);
	    } else {
		RB splash(80, 1, "FSM mode on");
		ADDTAG(cplayer, is_invulnerable);
	    }
	    break;
	case 4:
	    player_add_health(cplayer,100);
	    RB splash(80, 1, "Player touched by His Noodly Appendage...");
	    break;
	case 5:
	    if (HASTAG(cplayer, has_key)) {
		RB splash(80, 1, "Player doesn't have key");
		DELTAG(cplayer, has_key);
	    } else {
		RB splash(80, 1, "Player has key");
		ADDTAG(cplayer, has_key);
	    }
	    break;
            
	case MENU_ATTACHED_USB:
	    ret = 2;
	    break;
	default:
	    ret = 1;
	    break;
	}
    }

    restore_screen(screen_save);
    free(screen_save);
    menu_free(m);
    return ret;
}

int loadsave_menu(int load) {
    int ret = 0;
    int res = 0;
    char fname[MAX_PATH];
    char title[32];
    char disp[36];
    void* screen_save = save_screen();
    menu* m = new_menu(load ? "Load game" : "Save game", LEFT_EXITS);
    int i;

    if (load)
        menu_add_item(m, "Autosave", 1);
    for (i = 0; i < NUM_SAVE_GAMES; i++) {
        RB snprintf(fname, MAX_PATH, PATH "sg%d.sav", i);

        RB snprintf(disp, 36, "%d%s %s", i + 1, i<9?" ":"", 
                    save_game_exist[i] ? save_game_titles[i] : "<empty>");
        menu_add_item(m, disp, (!load || save_game_exist[i] ? i + 2 : 0));
    }

    res = show_menu(m);
    switch (res) {
    case 0:
        break;
    case MENU_ATTACHED_USB:
        ret = 3;
        break;
    default:
        if (res == 1) {
            RB strncpy(fname, PATH "auto.sav", MAX_PATH);
        } else {
            i = res - 2;
            RB snprintf(fname, MAX_PATH, PATH "sg%d.sav", i);
        }
        if (load) {
            music_change(NULL, MUS_FADE_SPD, 0, false);
            RB splash(0, 1, "Loading");
            if (load_game(fname)) {
                RB splash(150, 1, "Could not load game");
                music_change("intro", MUS_FADE_SPD, 0, true);
            } else {
                ret = 1;
            }
        } else {
            RB strncpy(title, save_game_titles[i], 32);
            RB lcd_set_drawmode(DRMODE_SOLID);
            RB lcd_set_foreground(0);
            if (!RB kbd_input(title, 32)) {
                RB splash(0, 1, "Saving");
                if (save_game(fname, title)) {
                    RB splash(150, 1, "Could not save game");
                } else {
                    RB strncpy(save_game_titles[i], title, 32);
                    save_game_exist[i] = 1;
                }
            }
            
        }
    }

    restore_screen(screen_save);
    free(screen_save);
    menu_free(m);
    return ret;
}

void play_example_sound(void) {
    static int holdoff = 0;
    if (*RB current_tick >= holdoff) {
        holdoff = *RB current_tick + (HZ/4);
        playsnd("pickup");
    }
    
}

void sound_volume_set(void* v, slider_data* dat, int dir) {
    (void)v;
    if (dir) {
        soundvol += dir*4;
        if (soundvol < 0) soundvol = 0;
        if (soundvol > 256) soundvol = 256;
        play_example_sound();
        mix_gametick();
    }
    dat->min = 0;
    dat->max = 256;
    dat->val = soundvol;
}

void music_volume_set(void* v, slider_data* dat, int dir) {
    (void)v;
    musicvol += dir * 4;
    if (musicvol > 256) musicvol = 256;
    if (musicvol < 0) musicvol = 0;

    dat->min = 0;
    dat->max = 256;
    dat->val = musicvol;
}

void system_volume_set(void* v, slider_data* dat, int dir) {
    (void)v;
    dat->min = RB sound_min(0);
    dat->max = RB sound_max(0);
    int step = (dat->max - dat->min) / 32;
    if (dir) {
        int newval = RB global_settings->volume + dir * step;
        if (newval < dat->min) newval = dat->min;
        if (newval > dat->max) newval = dat->max;
        RB global_settings->volume = newval;
        RB sound_set(0, newval);
        play_example_sound();
    }

    dat->val = RB global_settings->volume;
}

int settings_menu(void) {
    int ret = -1;
    int res;
    void* screen_save = save_screen();

    menu* m = new_menu("Settings", LEFT_EXITS);
    menu_add_item(m, "System volume", 1);
    menu_add_item(m, "Sound volume", 2);
    menu_add_item(m, "Music volume", 3);

    while (ret == -1) {
        restore_screen(screen_save);
        switch (show_menu(m)) {
        case 1:
            menudraw(m, false);
            res = slider("System volume", system_volume_set, NULL, false);
            if (res == 3) ret = 3;
            break;
        case 2:
            menudraw(m, false);
            res = slider("Sound volume", sound_volume_set, NULL, false);
            if (res == 3) ret = 3;
            break;
        case 3:
            menudraw(m, false);
            res = slider("Music volume", music_volume_set, NULL, false);
            if (res == 3) ret = 3;
            break;
        case MENU_ATTACHED_USB:
            ret = 3;
            break;
            
        default:
            ret = 0;
            break;
        }
    }
    restore_screen(screen_save);
    free(screen_save);
    menu_free(m);
    return ret;
}

int game_menu(void) {
    int ret = 0;
    void* screen_save = save_screen();

    menu* m = new_menu("Main Menu", (cplayer ? LEFT_EXITS : 0));
    if (cplayer) {
        menu_add_item(m, "Return to game", 2);
    } else {
        m->y = 40;
        menu_add_item(m, "New game", 1);
    }
    menu_add_item(m, "Settings", 7);
    menu_add_item(m, "Load game", 4);
    menu_add_item(m, "Save game", cplayer ? 3 : 0);
    menu_add_item(m, "End game", cplayer ? 5 : 0);
    menu_add_item(m, "Exit RocKen", 6);

    while (!ret) {
        restore_screen(screen_save);
        if (!cplayer) {
            introdraw(0, 0, 0, 2, -24576, false, false, true);
        }
        switch (show_menu(m)) {
        case 1: /* new game */
            new_game(NULL);
            ret = 1;
            break;

        case 7: /* settings*/
            menudraw(m, false);
            ret = settings_menu();
            break;

        case 2: /* return to game*/
            ret = 1;
            break;

        case 3: /* save game */
            menudraw(m, false);
            ret = loadsave_menu(0);
            break;

        case 4: /* load game */
            menudraw(m, false);
            ret = loadsave_menu(1);
            break;

        case 5: /* end game */
            end_game();
            music_change("intro", 8, 8, true);

            RB lcd_clear_display();
            ret = 1;
            break;

        case 6: /* exit */
            ret = 2;
            break;

        case MENU_ATTACHED_USB:
            ret = 3;
            break;

        default:
            if (cplayer)
                ret = 1;
            else
                ret = 2;
            break;
        }
    }
    free(screen_save);
    menu_free(m);
    RB lcd_clear_display();
    return ret;
}


void* memcpy(void* to, const void* from, size_t s) {
    return RB memcpy(to,from,s);
}

int rungame(void) {
    int button;
    int done = false;
    int ctime;
    int nexttick;
    timeprec_t end, start;

    int fvel, avel, svel, zvel;
    int fire;
    bool paused = false;


    action mainact = A_NONE;

    nexttick = *RB current_tick;
    while (!done) {
	cframe++;
	get_time_div(start);
#ifdef HAS_BUTTON_HOLD
        if (RB button_hold()) {
            if (!paused)
                music_fade(-MUS_FADE_SPD);
            paused = true;
        } else {
            if (paused)
                music_fade(MUS_FADE_SPD);
            paused = false;
        }
#endif

	if (paused) {
            RB lcd_setfont(FONT_UI);
            RB splash(0, 1, "Paused");
        } else {
            num_sprites = 0;

            move_actors();
            mix_gametick();
	
	

            board_changed();
	
            RB lcd_set_foreground(2);
            RB lcd_fillrect(0, 0, XDIM, YDIM/2);
            RB lcd_set_foreground(1);
            RB lcd_fillrect(0, YDIM/2, XDIM, YDIM/2);

            preparescreen(cplayer->x, cplayer->y, player_get_posz(cplayer), cplayer->ang, 0, YDIM, YDIM/2);

            drawscreen(RB lcd_framebuffer);

            draw_actors(); 

            RB yield();

            drawsprites(RB lcd_framebuffer);


            get_time_div(end);
            updatestatus(get_interval_usec(&start,&end));


	}

        RB lcd_update();
        
        if (HASTAG(cplayer, is_dying)) {
            music_cache("intro2");
            DELTAG(cplayer, is_dying);
        }

        if (HASTAG(cplayer, is_dead)) {
            music_change(NULL, MUS_FADE_SPD, 0, false);
            RB lcd_setfont(FONT_UI);
            RB splash(160, true, "Game Over");
            end_game();
            music_change("intro2", 8, 8, true);
            RB lcd_clear_display();
            return PLUGIN_OK;
        }
#if 1
	while ((ctime = *RB current_tick) < nexttick) {
	    /*RB sleep(nexttick - ctime);*/
            RB sleep(0);
	}
	nexttick += 8;
	if (nexttick < ctime) {
	    RB yield();
	    nexttick = ctime;
	}
#else
	RB yield();
#endif

	button = RB button_get_w_tmo(0);
	if (!(button & BUTTON_REMOTE)) {
	    if (button & BUTTON_REL) {
		mainact = A_NONE;
	    } else {
		switch (button) {
		case BUTTON_UP:
		    mainact = A_FWD;
		    break;
		case BUTTON_DOWN:
		    mainact = A_BACK;
		    break;
		case (BUTTON_ON|BUTTON_SELECT):
		    if (debugmenu() == 2) {
			return PLUGIN_USB_CONNECTED;
		    }
		    break;
		case (BUTTON_UP|BUTTON_ON):
		    mainact = A_UP;
		    break;
		case (BUTTON_DOWN|BUTTON_ON):
		    mainact = A_DOWN;
		    break;
		case BUTTON_LEFT:
		    mainact = A_LEFT;
		    break;
		case BUTTON_RIGHT:
		    mainact = A_RIGHT;
		    break;
		case BUTTON_REC:
		    mainact = A_SLEFT;
		    break;
		case BUTTON_MODE:
		    mainact = A_SRIGHT;
		    break;
		case BUTTON_SELECT:
		    mainact = A_FIRE;
		    break;
		case BUTTON_OFF:
		    done = true;
		    break;

		default:
                    if (default_button(button) == SYS_USB_CONNECTED) {
                        return PLUGIN_USB_CONNECTED;
                    }
		    break;
		}
	    }
	}
	fvel = svel = avel =  0;
	zvel = fire = 0;

	switch (mainact) {
	case A_NONE:
	    break;
	case A_FWD:
	    fvel = 1;
	    break;
	case A_BACK:
	    fvel = -1;
	    break;
	case A_LEFT:
	    avel = -1;
	    break;
	case A_RIGHT:
	    avel = 1;
	    break;
	case A_SLEFT:
	    svel = -1;
	    break;
	case A_SRIGHT:
	    svel = 1;
	    break;
	case A_UP:
	    zvel = -1;
	    break;
	case A_DOWN:
	    zvel = 1;
	    break;
	case A_FIRE:
	    fire = 1;
	    break;
	}
	player_set_controls(cplayer, fvel, svel, zvel, avel, fire);
    }
    return PLUGIN_OK;
}


#ifdef USE_IRAM
extern char iramcopy[];
extern char iramstart[];
extern char iramend[];
extern char iedata[];
extern char iend[];
#endif

typedef void(*ctor_func)(void);

/* defined in linker script */
extern ctor_func _ctors_start[];
extern ctor_func _ctors_end[];

enum plugin_status plugin_start(struct plugin_api *api, void *parameter) {
    int res, i;
    uint8 *mallocbuf;
    int mallocbuf_size = 0;

    (void)parameter;

    rb = api;

    /* this has to be done before we copy iram data */

#ifndef SIMULATOR
    mallocbuf = api->plugin_get_audio_buffer(&mallocbuf_size);
#endif

#ifdef USE_IRAM
    api->memcpy(iramstart, iramcopy, iramend - iramstart);
    api->memset(iedata, 0, iend - iedata);
#endif

    /* copy plugin api to iram for speed */
    api->memcpy(&rb_api, api, sizeof(rb_api));

#ifndef SIMULATOR
    malloc_init(mallocbuf, mallocbuf_size);

    /* Run our constructors. On the simulator, these have already been
       called. */
    ctor_func* cctor;
    for (cctor = _ctors_start; cctor < _ctors_end; cctor++) {
	(*cctor)();
    }
#endif

    load_settings(PATH "rocken.opt");

    sound_init();

    load_sounds(PATH "sounds");

    for (i = 0; i < NUM_SAVE_GAMES; i++) {
        char fname[MAX_PATH];
        RB snprintf(fname, MAX_PATH, PATH "sg%d.sav", i);

        if (get_savegame_title(fname, save_game_titles[i])) {
            RB strncpy(save_game_titles[i], "", 32);
            save_game_exist[i] = 0;
        } else {
            save_game_exist[i] = 1;
        }
    }

    RB lcd_remote_setfont(FONT_SYSFIXED);
    RB lcd_setfont(FONT_SYSFIXED);

    RB lcd_clear_display();

    res = loadtex(PATH "rocken.tex");
    if (res) {
	RB splash(150, 1, "Could not load textures: %d", res);
        game_shutdown();
        sound_fini();
	return PLUGIN_OK;
    }

#ifndef SIMULATOR
    RB cpu_boost(1);
#endif

    if (parameter) {
        char* fname = (char *)parameter;
        char* ext = RB strrchr(fname, '.');
        if (ext && !RB strcmp(ext,".rbd")) {
            res = loadboard(fname);
        } else {
            res = load_game(fname);
        }
        if (res) {
            RB splash(150, 1, "Could not load %s: %d", fname, res);
            game_shutdown();
            sound_fini();
            return PLUGIN_OK;
        }
        res = 1;
    } else {
        music_change("begin", 0, 0, false);
        music_wait_load();
        music_cache("intro");
        RB sleep(5);
        res = intro();
        RB lcd_clear_display();

        music_uncache("begin");
        music_change("intro", 10, 10, true);
    }
    while (res == 1) {
        if (cplayer) {
            res = rungame();
            if (res == PLUGIN_USB_CONNECTED) {
                res = 3;
                break;
            }
        }
        res = game_menu();
    }

    game_shutdown();
    sound_fini();

    return res == 3 ? PLUGIN_USB_CONNECTED : PLUGIN_OK;
}
