/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 *
 * Copyright (C) 2006 Katie Stafford
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#ifndef __TAG_H_
#define __TAG_H_

#include "common.h"
#include "datafile.h"

#define TAG_INVALID 0

typedef struct {
    union {
        uint32 shdata[4];
        struct {
            uint32 flags;
            uint32 len;
            uint32* data;
        };
    };
} tagset;

typedef uint32 tag;

/* Functions which are used interally by inline functions */
extern tag _get_tag(const char* name, int dup);
extern void _tagset_copy_long(tagset* to, const tagset* from);
extern void _tagset_clear_long(tagset* ts);
extern void _tagset_add_long(tagset* ts, tag t);
extern void _tagset_del_long(tagset* ts, tag t);
extern int _tagset_in_long(const tagset* ts, tag t);
extern void _tagset_compose(tagset* ts, const char* lst);

/* Return the name of a tag. */
extern const char* tag_name(tag t);

/* Calls _func_ for every tag in _ts_. */
extern void tagset_enumerate(const tagset* ts, void (*func)(tag, void*), void*);

/* Create a list of all the tags in _ts_. The caller is responsible
   for freeing _tags_. */
extern void tagset_list(const tagset* ts, tag** tags, uint32* ntags);

/* Read or write a tagset to a buffer. */
extern void tagset_write(buffer* buf, tagset* ts);
extern void tagset_read(buffer* buf, tagset* ts);

void serial_tagset(tagset* ts, buffer* buf, int phase);

/* Returns 1 if _t_ is in _ts_. */
static inline int tagset_in(const tagset* ts, tag t) {
    if (ts->flags & 1)
        return _tagset_in_long(ts, t);
    if (t > 127)
        return 0;
    return (ts->shdata[t>>5]>>(t&31)) & 1;
}

/* Add tag _t_ to tagset _ts_. */
static inline void tagset_add(tagset* ts, tag t) {
    if (t > 127 || ts->flags & 1) {
        _tagset_add_long(ts, t);
        return;
    }
    ts->shdata[t>>5] |= 1<<(t&31);
}

/* Remove tag _t_ from tagset _ts_. */
static inline void tagset_del(tagset* ts, tag t) {
    if (ts->flags & 1) {
        _tagset_del_long(ts, t);
        return;
    }
    if (t > 127)
        return;
    ts->shdata[t>>5] &= ~(1<<(t&31));
}

/* Copy tagset _from_ to tagset _to_. */
static inline void tagset_copy(tagset* to, const tagset* from) {
    if (from->flags & 1) {
        _tagset_copy_long(to, from);
        return;
    }
    to->shdata[0] = from->shdata[0];
    to->shdata[1] = from->shdata[1];
    to->shdata[2] = from->shdata[2];
    to->shdata[3] = from->shdata[3];
}

/* Clear tagset _ts_. This should not be used to initialize a tagset. */
static inline void tagset_clear(tagset* ts) {
    if (ts->flags & 1) {
        _tagset_clear_long(ts);
        return;
    }
    ts->shdata[0] = ts->shdata[1] = 
        ts->shdata[2] = ts->shdata[3] = 0;
}

/* Initialize tagset _ts_. This should not be used to clear a
   tagset (it may leak memory). */
static inline void tagset_init(tagset* ts) {
    ts->shdata[0] = ts->shdata[1] = 
        ts->shdata[2] = ts->shdata[3] = 0;
}

/* Lookup a tag in the table. Adds a copy if it is not already in the
   table. */
#define get_tag(name) _get_tag(name, 1)

#define _TAG(nm)                                \
({                                              \
    static tag _t = 0;                          \
    if (!_t) {                                  \
        _t = _get_tag(nm, 0);                   \
    }                                           \
    _t;                                         \
})

/* Lookup a tag in the table and cache the result. */
#define TAG(nm) _TAG(#nm)

#define _TAGSET(lst)                            \
({                                              \
    static tagset _ts;                          \
    static char _init;                          \
    if (!_init) {                               \
        _init = 1;                              \
        _tagset_compose(&_ts, lst);             \
    }                                           \
    &_ts;                                       \
})

/* Initialize a tagset with the given tags. */
#define TAGSET(...) _TAGSET(#__VA_ARGS__)

#define TAGSET_STATIC_INIT(ts, ...)             \
({                                              \
    static char _init;                          \
    if (!_init) {                               \
        _init = 1;                              \
        _tagset_compose((&ts), #__VA_ARGS__);   \
    }                                           \
})

#define TAGSET_ADD(ts, nm) tagset_add(ts, _TAG(#nm))
#define TAGSET_DEL(ts, nm) tagset_del(ts, _TAG(#nm))
#define TAGSET_IN(ts, nm) tagset_in(ts, _TAG(#nm))

#define TAGSET_INIT {{{0, 0, 0, 0}}}

#endif
